#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "usbdrv.h"
#include "messages.h"

//handy macros for laziness
//not technically cbi and sbi in most cases, but the name works
#define cbi(reg, bit) (reg) &= ~(1 << (bit))
#define sbi(reg, bit) (reg) |= (1 << (bit))


//control via DDR (hi-z/low)
#define RST_DOWN() sbi(DDRD, 4)
#define RST_UP() cbi(DDRD, 4)

//TMS/BKPT = C4
#define BKPT_DOWN() sbi(DDRC, 4)
#define BKPT_UP() cbi(DDRC, 4)

//control via port (hi/low)
#define TDI_UP() sbi(PORTD, 0)
#define TDI_DOWN() cbi(PORTD, 0)

//DSCLK (TRST) = C5
#define CLK_UP() sbi(PORTC, 5)
#define CLK_DOWN() cbi(PORTC, 5)

#define TDO (PIND & (1 << 1))

//Include the initial data to be written to the chip on bootup
//having this include the code to make the H120 stay on is a good idea
//stored in PROGMEM, so it can use up all of the chip's flash if required
//#include "init.c"
#include "../h120_init/init_prog.c"


/* ---- BDM code first ---- */
//see the MCF5249 or CF2UM docs for details

uchar xfer_inbuf[2];
uchar xfer_outbuf[2];

//shift the data out and the return value in
//the 3us delays may or may not be necessary, but they don't hurt
uchar xfer_8(uchar data) {
	int i;
	for (i = 0;i<8;i++) {
		if (data & 0x80)
			TDI_UP();
		else
			TDI_DOWN();
		data <<= 1;
		_delay_us(3);
		CLK_UP();
		_delay_us(3);
		CLK_DOWN();
		if (TDO)
			data |= 1;
	}
	return data;
}

uchar xfer() {
	uchar out = 0;
	//do the 1st bit
	TDI_DOWN();
	CLK_UP();
	_delay_us(1);
	CLK_DOWN();
	if (TDO)
		out = 1;

	xfer_outbuf[0] = xfer_8(xfer_inbuf[0]);
	xfer_outbuf[1] = xfer_8(xfer_inbuf[1]);
	return out;
}

//Resync the transmission, in case a bit gets lost
void bdm_sync() {

	//write out 2 commands worth of 0s
	TDI_DOWN();

	int i;
	for (i = 0; i < 34;i++) {
		CLK_UP();
		_delay_us(1);
		CLK_DOWN();
		_delay_us(1);
	}

	//keep writing 0s until the return is a 0
	
	while(TDO) {
		CLK_UP();
		_delay_us(1);
		CLK_DOWN();
		_delay_us(1);
	}

	//return was a 0, move the rest of the 16 bits
	for (i = 0;i < 16;i++) {
		CLK_UP();
		_delay_us(1);
		CLK_DOWN();
		_delay_us(1);
	}

}

// GO command, resume execution
void bdm_go() {
	xfer_inbuf[0] = 0xc;
	xfer_inbuf[1] = 0;
	xfer();
	return;
}

//put data to write here. Data to be read will also go in here.
uchar bdm_databuf[8];

//do nothing, good for wait loops
uchar bdm_nop() {
	xfer_inbuf[0] = 0;
	xfer_inbuf[1] = 0;
	return xfer();
}

//wait for a blank return value (0.ffff)
void bdm_wait() {
	uchar out;
	while(1) {
		out = bdm_nop();
		if (!out && xfer_outbuf[0] == 0xff && xfer_outbuf[1] == 0xff)
			break;
	}
	return;
}

//write out the given number of bytes from the databuf to the chip
//everything here assumes transfers will be done as pairs of bytes
void bdm_writebuf(uchar count) {
	uchar *ptr = bdm_databuf;
	while(count > 0) {
		xfer_inbuf[0] = *ptr++;
		xfer_inbuf[1] = *ptr++;
		xfer();

		count -= 2;
	}
	return;
}

//write 1 word to the chip
//format: address address address address data data
void bdm_write16() {
	//command
	xfer_inbuf[0] = 0x18;
	xfer_inbuf[1] = 0x40;
	xfer();
	bdm_writebuf(6);

	xfer_inbuf[0] = 0;
	xfer_inbuf[1] = 0;
	while(xfer()); //wait for transfer to finish

	return;
}

//read 1 word from the chip
//format: address address address address
//return: read read
void bdm_read16() {
	xfer_inbuf[0] = 0x19;
	xfer_inbuf[1] = 0x40;
	xfer();
	bdm_writebuf(4);
	xfer_inbuf[0] = 0;
	xfer_inbuf[1] = 0;
	while(xfer());
	// outbuf is the returned value
	bdm_databuf[0] = xfer_outbuf[0];
	bdm_databuf[1] = xfer_outbuf[1];
	return;
}

//Fill in a control register
//format: reg reg data data data data
void bdm_writecr() {
	//command
	xfer_inbuf[0] = 0x28;
	xfer_inbuf[1] = 0x80;
	xfer();
	//high word of reg (always 0)
	xfer_inbuf[0] = 0;
	xfer_inbuf[1] = 0;
	xfer();
	//low word of reg
	xfer_inbuf[0] = bdm_databuf[0];
	xfer_inbuf[1] = bdm_databuf[1];
	xfer();

	//data to write to reg
	xfer_inbuf[0] = bdm_databuf[2];
	xfer_inbuf[1] = bdm_databuf[3];
	xfer();
	xfer_inbuf[0] = bdm_databuf[4];
	xfer_inbuf[1] = bdm_databuf[5];
	xfer();
	
	xfer_inbuf[0] = 0;
	xfer_inbuf[1] = 0;
	while(xfer());
	return;
}

// read the processor status nibble
uchar pst_read() {
	uchar out = 0;
	if (PIND & (1 << 5))
		out |= 0x8;
	if (PIND & (1 << 6))
		out |= 0x4;
	if (PIND & (1 << 7))
		out |= 0x2;
	if (PINB & (1 << 0))
		out |= 0x1;
	return out;
}

// write something to the 4 PST leds
// so luxurious having 4 bits of debug output...
void pst_display(uchar pst) {
	if (pst & 1)
		sbi(PORTC, 0);
	else
		cbi(PORTC, 0);

	if (pst & 2)
		sbi(PORTC, 1);
	else
		cbi(PORTC, 1);

	if (pst & 4)
		sbi(PORTC, 2);
	else
		cbi(PORTC, 2);

	if (pst & 8)
		sbi(PORTC, 3);
	else
		cbi(PORTC, 3);

	return;
}

// reset the chip into HALT mode
void reset_bdm() {
	//bkpt low, pulse RST low
	BKPT_DOWN();
	RST_DOWN();
	CLK_DOWN();
	TDI_DOWN();
	_delay_ms(50);
	RST_UP();
	_delay_ms(80);
	BKPT_UP();
	bdm_nop();
	return;
}

// just halt the chip
void bdm_halt() {
	//bkpt low
	BKPT_DOWN();
	_delay_ms(80);
	BKPT_UP();
	bdm_nop();
	return;
}

/* ---- USB stuff ---- */
//USB driver is obdev's V-USB

uchar buf_used;
uchar xfer_complete = 0;
uchar status;
#define BUFSIZE 16 // how much space can I afford?
uchar cbuf[BUFSIZE];
uchar *wp;
unsigned short xfersize;
uchar recvd;

uchar cmd;

uchar usbFunctionSetup(uchar data[8]) {
	usbRequest_t *rq = data;

	static uchar reply[8];
	usbMsgPtr = reply;

	int i;
	switch(rq->bRequest) {
		case MSG_STATUS:
			reply[0] = status;
			return 1;
		case MSG_EOF: 
			reply[0] = MSG_WAIT;
			status = MSG_WAIT;
			xfer_complete = 1;
			return 1;
		case MSG_ECHO:
			//for testing
			reply[0] = data[2]; //2 bytes of the value field
			reply[1] = data[3];
			return 2;
		case MSG_OUT:
			for (i = 0;i < 8;i++) {
				reply[i] = cbuf[i];
			}
			return 8;
		case MSG_WRITE:
			//after this returns the rest of the read goes to
			//usbFunctionWrite
			xfersize = (unsigned short)rq->wLength.word;
			wp = cbuf;
			recvd = 0;
			return USB_NO_MSG;
		case MSG_RELOAD:
			cmd = MSG_RELOAD;
			reply[0] = 0;
			return 1;

		case MSG_RESET:
			cmd = MSG_RESET;
			return 0;

		case MSG_READ:
			//this might mess up other transfers
			bdm_databuf[0] = 0;
			bdm_databuf[1] = 0;
			bdm_databuf[2] = data[3];
			bdm_databuf[3] = data[2]; //possibly endianness issues again...
			bdm_read16();
			reply[0] = bdm_databuf[0];
			reply[1] = bdm_databuf[1];
			return 2;
			
		default:
			//not like the PC code actually cares
			reply[0] = MSG_ERR;
			return 1;
	}

	return 0; //shouldn't reach here...
}

//The _only_ control out accesses should be writing data
uchar usbFunctionWrite(uchar *data, uchar len) {
	uchar i;
	for (i = 0;i < len;i++)
		*wp++ = *data++;
	status = MSG_WAIT;
	recvd += len;

	if (recvd == xfersize) {
		buf_used = recvd;
		return 1;
	}
	return 0;
}

int main() {

// Port D: 00000001
	DDRD = 0x01;
// Port C: 00101111
	DDRC = 0x2F;

	usbInit();
	sei();

	//halt the chip
	reset_bdm();

	//this needs to be executed fairly frequently, so keep running it
	//whenever
	usbPoll();

	//wait for the chip to halt
	uchar pst = pst_read();
	while (pst != 0xf) {
		usbPoll();
		pst_display(pst);
		reset_bdm();
		pst = pst_read();
	}
	
	//transmission should be synced already, but just in case...
	bdm_sync();

	usbPoll();


	//load 0x2700 into SR (supervisor mode and interrupts disable)
	bdm_databuf[0] = 0x8;
	bdm_databuf[1] = 0x0E;

	bdm_databuf[2] = 0x00;
	bdm_databuf[3] = 0x00;
	bdm_databuf[4] = 0x27;
	bdm_databuf[5] = 0x00;
	bdm_writecr();
	
	//first set the RAMBAR registers
	//RAMBAR0 = 0x10010001
	//32K dma-incapable sram
	//placed at 0x10010000
	
	bdm_databuf[0] = 0xC; //C04 = RAMBAR0
	bdm_databuf[1] = 0x04;
	bdm_databuf[2] = 0x10;
	bdm_databuf[3] = 0x01;
	bdm_databuf[4] = 0x00;
	bdm_databuf[5] = 0x01;
	bdm_writecr();
	
	//RAMBAR1 = #0x10000e01
	//64K dma-capable sram
	//placed at 0x10000000

	bdm_databuf[0] = 0xC; //C05 = RAMBAR1
	bdm_databuf[1] = 0x05;
	bdm_databuf[2] = 0x10;
	bdm_databuf[3] = 0x00;
	bdm_databuf[4] = 0x0e;
	bdm_databuf[5] = 0x01;
	bdm_writecr();

	// init completed. Write out the bootloader thingo

	//write a bit of code to start of program ram (0x10000000)
	bdm_databuf[0] = 0x10;
	bdm_databuf[1] = 0;
	bdm_databuf[2] = 0;
	bdm_databuf[3] = 0;

	unsigned short i;

	union {
		uint16_t w;
		uchar c[2];
	} offset;
	offset.w = 0;

	for (i = 0;i < PROGSIZE; i+= 2) {
		bdm_databuf[2] = offset.c[1]; //stupid endianness problems
		bdm_databuf[3] = offset.c[0];
	
	//pgm_read_byte because it's stored in flash (only 1K of ram...)
		bdm_databuf[4] = pgm_read_byte(&prog[i]);
		bdm_databuf[5] = pgm_read_byte(&prog[i + 1]);
		bdm_write16();

		offset.w += 2;
		usbPoll();
	}

	//set PC to start of code
	bdm_databuf[0] = 0x8; //80F = PC
	bdm_databuf[1] = 0x0F;
	bdm_databuf[2] = 0x10;
	bdm_databuf[3] = 0x00;
	bdm_databuf[4] = 0x00;
	bdm_databuf[5] = 0x00;
	bdm_writecr();

	//read the data back to verify
	bdm_databuf[0] = 0x10;
	bdm_databuf[1] = 0;
	bdm_databuf[2] = 0;
	bdm_databuf[3] = 0;

	offset.w = 0;
	//this is mainly to identify communication errors
	for (i = 0;i < PROGSIZE; i+= 2) {
		bdm_databuf[2] = offset.c[1];
		bdm_databuf[3] = offset.c[0];

		bdm_read16();
		uchar buf[2];
		buf[0] = pgm_read_byte(&prog[i]);
		buf[1] = pgm_read_byte(&prog[i + 1]);

		if (buf[0] != bdm_databuf[0] ||
				buf[1] != bdm_databuf[1]) {
			pst_display(0x6); //reserved value, good for errors
			while(1);
			//very bad if the code ever gets here - the transfer
			//failed.
		}

		//reset these in case a USB read messes them up
		bdm_databuf[0] = 0x10;
		bdm_databuf[1] = 0;
		
		offset.w += 2;

		usbPoll();
	}
	//all good, execute the code

	bdm_go();

	//from here, the code reads in data from USB, writes it to 0x10000000,
	//and executes it when an EOF arrives
reload:

	while(pst != 0xf) {
		//wait for the CPU to halt again
		usbPoll();
		pst = pst_read();
		pst_display(pst);
	}

	cmd = 0;

	status = MSG_SEND;
	bdm_databuf[0] = 0x10;
	bdm_databuf[1] = 0;
	offset.w = 0;
	xfer_complete = 0;

	//receive data via the USB control reads, write them out to memory
	while(1) {
		usbPoll();
		pst = pst_read();
		pst_display(pst);
		if (buf_used) {
			//MSG_READ rewrites these
			bdm_databuf[0] = 0x10;
			bdm_databuf[1] = 0;
			for (i = 0;i < buf_used;i+=2) {
				bdm_databuf[2] = offset.c[1];
				bdm_databuf[3] = offset.c[0];
				bdm_databuf[4] = cbuf[i];
				bdm_databuf[5] = cbuf[i+1];
				bdm_write16();
				offset.w += 2;
				usbPoll();
				pst_display(pst_read());
			}
			buf_used = 0;
			status = MSG_SEND;
		}
		//finished if buffer is empty and xfer_complete is set (received
		//EOF)
		if (buf_used == 0 && xfer_complete)
			break;
	}

	//write the PC again
	bdm_databuf[0] = 0x8; //80F = PC
	bdm_databuf[1] = 0x0F;
	bdm_databuf[2] = 0x10;
	bdm_databuf[3] = 0x00;
	bdm_databuf[4] = 0x00;
	bdm_databuf[5] = 0x00;
	bdm_writecr();


	bdm_go();

	//keep displaying status and watch for new commands

	cmd = 0;
	while(1) {
		usbPoll();
		pst_display(pst_read());
		if (cmd == MSG_RESET) {
			reset_bdm();
			cmd = 0;
		}

		if (cmd == MSG_RELOAD) {
			bdm_halt();
			cmd = 0;
			goto reload;
		}
	}

	//won't reach here
	return 0;
}

